// ============================================================================================
// ===																						===
// === phreshold.c																			===
// ===																						===
// === Functions defined:																	===
// ===		load_bitmap																		===
// ===		phreshold_init																	===
// ===		refresh_frame																	===
// ===		phreshold_setup																	===
// ===		phreshold_destroyed																===
// ===		draw_score																		===
// ===		draw_frame																		===
// ===		freeze																			===
// ===		unfreeze																		===
// ===																						===
// ============================================================================================

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <limits.h>
#include <time.h>
#include <sys/proxy.h>
#include <sys/kernel.h>
#include <sys/psinfo.h>
#include <sys/osinfo.h>
#include <sys/sched.h>
#include <process.h>
#include <sys/qnx_glob.h>
#include <semaphore.h>
#include <pwd.h>
#include <stdarg.h>

/* Toolkit headers */
#include <Ph.h>
#include <Pt.h>
#include <Ap.h>

#define PX_IMAGE_MODULES
#define PX_BMP_SUPPORT
#include <photon/PxImage.h>

/* Local headers */
#define __MAIN__
#include "phreshold.h"
#include "abimport.h"
#include "proto.h"


/*--- Common definitions shared by the two tasks: phreshold and phreshold_kbd ---*/
#include "phreshold_kbd.h"
KEYS_SHARED		*keys_state;

/* Application Options string */
const char ApOptions[] =
	AB_OPTIONS ""; /* Add your options in the "" */


/*--- Used by:	set-exe_dir() and restor_previous_dir ---*/
static char		__currdir[ PATH_MAX + 1 ];


static void set_exe_dir( void )
	{
	char		path[ PATH_MAX + 1 ], fulldir[ PATH_MAX + 1 ];
	char		_node[ _MAX_NODE + 1 ], _dir[ _MAX_DIR + 1 ],
				_fname[ _MAX_FNAME + 1 ], _ext[ _MAX_EXT + 1 ];

	/*--- Set the current directory where the executable was loaded ---*/
	memset( __currdir, 0, sizeof( __currdir ) );
	if ( getcwd( __currdir, PATH_MAX ) == NULL )
		return;
	if ( _cmdname( path ) == NULL )
		return;
	_splitpath( path, _node, _dir, _fname, _ext );
	sprintf( fulldir, "%s%s", _node, _dir );
	chdir( fulldir );

	}	// set_exe_dir


static void restore_previous_dir( void )
	{

	/*--- Restore the original current directory ---*/
	chdir( __currdir );
	
	}	// restore_previous_dir


PhImage_t	*load_bitmap( const char *bitmap_name )
	{
	PhImage_t		*load;
	PxMethods_t		methods;
	char			fname[ 100 ];

	/*--- Goto into the directory where executable was found ---*/
	set_exe_dir();

	/*--- Get the bitmap ---*/
	memset( &methods, 0, sizeof( methods ) );
	methods.flags = PX_LOAD;
	sprintf( fname, "bitmaps/%s", bitmap_name );
	if ( (load = PxLoadImage( fname, &methods )) == NULL )
		{
		fprintf( stderr, "Unable to load bitmap '%s'\n", fname );
		PtDestroyWidget( ABW_phreshold );
		exit( -1 );
		}

	/*--- Restore previous current directory ---*/
	restore_previous_dir();

	/*--- Ok ---*/
	return load;

	}	// load_bitmap


/*--- ---------------------------------------------- ---*/
/*--- Some global initializations (once, that's all) ---*/
/*--- ---------------------------------------------- ---*/

static void init_first( void )
	{
	STARTING_CANNON		*starting_cannon;
	int					n;
	PtArg_t				args[ 1 ];
	PhArea_t			*area;
	char				fname[ 80 ];

	/*--- Get the positions of the starting cannons ---*/
	for ( starting_cannon = starting_cannons, n = 0; 
		  n < NB_STARTING_CANNONS; 
		  n++, starting_cannon++ )
		{
		starting_cannon->wgt = dsphreshold[ ABN_phreshold_start_cannon_0 + n ].wgt;
		PtSetArg( &args[ 0 ], Pt_ARG_AREA, &area, 0 );
		PtGetResources( dsphreshold[ ABN_phreshold_start_cannon_0 + n ].wgt, 1, args );
		starting_cannon->init_x = area->pos.x;
		starting_cannon->init_y = area->pos.y;
		starting_cannon->w = area->size.w;
		starting_cannon->h = area->size.h;
		}	// for (n, starting_cannon)

	/*--- Get the bitmap of the shell ---*/
	shell_img_des = load_bitmap( "shell.0.bmp" );
	shell_img = shell_img_des->image;
	shell_w   = shell_img_des->size.w;
	shell_h   = shell_img_des->size.h;

	/*--- Get the bitmaps of the cannon ---*/
	cannon_img_des[ CANNON_ARMED_0 ]   = load_bitmap( "cannon.0.bmp" );
	cannon_img_des[ CANNON_ARMED_1 ]   = load_bitmap( "cannon.1.bmp" );
	cannon_img_des[ CANNON_ARMED_2 ]   = load_bitmap( "cannon.2.bmp" );
	cannon_img_des[ CANNON_UNARMED ] = load_bitmap( "cannon.unarmed.bmp" );
	cannon_img_des[ CANNON_DEAD ]    = load_bitmap( "cannon.dead.bmp" );
	cannon_img[ CANNON_ARMED_0 ]   = cannon_img_des[ CANNON_ARMED_0 ]->image;
	cannon_img[ CANNON_ARMED_1 ]   = cannon_img_des[ CANNON_ARMED_1 ]->image;
	cannon_img[ CANNON_ARMED_2 ]   = cannon_img_des[ CANNON_ARMED_2 ]->image;
	cannon_img[ CANNON_UNARMED ] = cannon_img_des[ CANNON_UNARMED ]->image;
	cannon_img[ CANNON_DEAD ]    = cannon_img_des[ CANNON_DEAD ]->image;
	cannon.w = cannon_img_des[ CANNON_ARMED_0 ]->size.w;
	cannon.h = cannon_img_des[ CANNON_ARMED_0 ]->size.h;

	/*--- Get the bitmaps of the missile ---*/
	missile_img_des = load_bitmap( "missile.0.bmp" );
	missile_img = missile_img_des->image;
	missile_w   = missile_img_des->size.w;
	missile_h   = missile_img_des->size.h;

	/*--- Get the bitmap for the mother ---*/
	mother_img_des = load_bitmap( "mother.bmp" );
	mother_img = mother_img_des->image;
	mother.w   = mother_img_des->size.w;
	mother.h   = mother_img_des->size.h;

	/*--- Get the bitmaps for the stars ---*/
	for ( n = 0; n < STAR_NBSTATES; n++ )
		{
		sprintf( fname, "star.%d.bmp", n );
		star_img_des[ n ] = load_bitmap( fname );
		star_img[ n ] = star_img_des[ n ]->image;
		}

	crc_palette = PxCRC( (char*)palette, 256 * sizeof( PgColor_t ) );
	memmove( palette,
			cannon_img_des[ 0 ]->palette, 
			sizeof( palette ) );

	}	// init_first
	

static void cannon_is_dead( void )
	{

	cannon.hitcnt = 0;
	if ( nb_cannons_left > 0 )
		{
		restart_same_level( TRUE );
		refresh_frame( TRUE );
		}

	}	// cannon_is_dead
	

void set_min_size( void )
	{

	frame_width  = FRAME_WIDTH_MIN;
	frame_height = FRAME_HEIGHT_MIN;
	frame_size = frame_width * frame_height;
	
	frame_dim.x = frame_width;
	frame_dim.y = frame_height;
	
	/*--- Allocate space for background ---*/
	if ( background )
		free( background );
	if ( (background = malloc( frame_size )) == NULL )
		{
		fprintf( stderr, "No memory.\n" );
		exit( -1 );
		}

	}	// set_min_size


void set_max_size( void )
	{

	frame_width  = FRAME_WIDTH_MAX;
	frame_height = FRAME_HEIGHT_MAX;
	frame_size = frame_width * frame_height;

	frame_dim.x = frame_width;
	frame_dim.y = frame_height;
	
	/*--- Allocate space for background ---*/
	if ( background )
		free( background );
	if ( (background = malloc( frame_size )) == NULL )
		{
		fprintf( stderr, "No memory.\n" );
		exit( -1 );
		}

	}	// set_max_size


static void create_kbd_task( void )
	{
	char		path[ PATH_MAX + 1 ], exename[ PATH_MAX + 1 ];
	char		_node[ _MAX_NODE + 1 ], _dir[ _MAX_DIR + 1 ],
				_fname[ _MAX_FNAME + 1 ], _ext[ _MAX_EXT + 1 ];

	/*--- Start the process 'phreshold_kbd' ---*/
	if ( _cmdname( path ) == NULL )
		{
		fprintf( stderr, "_cmdname failed\n" );
		exit( -1 );
		}
	_splitpath( path, _node, _dir, _fname, _ext );
	qnx_spawn_options.flags |= _SPAWN_NOZOMBIE;
	qnx_spawn_options.sched_algo = SCHED_RR;		// Round robin, not Fifo
	sprintf( exename, "%s%s/phreshold_kbd", _node, _dir );
	if ( pid_kbd != -1 )
		kill( pid_kbd, SIGTERM );
	if ( (pid_kbd = spawnl( P_NOWAIT, exename, exename, NULL )) == -1 )
		{
		fprintf( stderr, "Unable to run the process '%s'\n", exename );
		exit( -1 );
		}

	/*--- Create the shared memory ---*/
	if ( (work_frame = PgShmemCreate( frame_size, WORK_FRAME_SHNAME )) == NULL )
		{
		if ( (shm_unlink( WORK_FRAME_SHNAME )) == -1 )
			{
			fprintf( stderr, "Unable to unlink shared memory '%s' [err=%d]\n",
				WORK_FRAME_SHNAME, errno );
			exit( -1 );
			}
		if ( (work_frame = PgShmemCreate( frame_size, WORK_FRAME_SHNAME )) == NULL )
			{
			fprintf( stderr, "Unable to create shared memory '%s' [err=%d]\n",
				WORK_FRAME_SHNAME, errno );
			exit( -1 );
			}
		}	// if

	}	// create_kbd_task


static void extend_phreshold_size( void )
	{
	int				incx = 160;
	int				incy = 120;

	set_max_size();

	setByDim( ABW_phreshold, incx, incy );
	setByDim( ABW_phreshold_around, incx, incy );
	moveBy( ABW_phreshold_laser, incx, 0 );
	moveBy( ABW_phreshold_laser_label, incx, 0 );
	moveBy( ABW_phreshold_fuel, incx, 0 );
	moveBy( ABW_phreshold_fuel_label, incx, 0 );
	moveBy( ABW_phreshold_ramp, incx, 0 );
	moveBy( ABW_phreshold_start_cannon_0, incx, 0 );
	moveBy( ABW_phreshold_start_cannon_1, incx, 0 );
	moveBy( ABW_phreshold_start_cannon_2, incx, 0 );
	moveBy( ABW_phreshold_start_cannon_3, incx, 0 );
	moveBy( ABW_phreshold_start_cannon_4, incx, 0 );
	moveBy( ABW_phreshold_score, incx, incy );
	moveBy( ABW_phreshold_level, incx, incy );
	moveBy( ABW_phreshold_msg, 0, incy );

	}	// extend_phreshold_size


#if LOGTOTMP==1
void slog( const char *fmt, ... )
	{
	va_list			ap;
	char			msg[ 300 ];
	FILE			*fp;
	time_t			now;
	struct tm		*tm;

	/*--- Develop text to log ---*/
	va_start( ap, fmt );
	vsprintf( msg, fmt, ap );
	va_end( ap );

	/*--- Update log file ---*/
	time( &now );
	tm = localtime( &now );
	if ( (fp = fopen( "/tmp/log.txt", "a" )) != NULL )
		{
		fprintf( fp, "%02d/%02d %02d:%02d:%02d [phothello] [%s] %s\n",
			tm->tm_mday, tm->tm_mon+1,
			tm->tm_hour, tm->tm_min, tm->tm_sec,
			getlogin(),
			msg );
		fclose( fp );
		}

	}	// slog
#endif


int
phreshold_init( int argc, char *argv[] )
	{
	struct sigevent	sigevnt;
	PhSysInfo_t		sysinfo;
	int				n;
	int				fd;
	PhEvent_t		*event;
	pid_t			pid;
	int				cpt;
	int				nb_frames;
	clock_t			t_start;
	struct _osinfo	osdata;
	PhRid_t			rid;
	struct passwd	*pw;
int	max_nbframes_persec;

	/* eliminate 'unreferenced' warnings */
	argc = argc, argv = argv;

#if LOGTOTMP==1
	slog( "start" );
#endif

	/*--- Buffer for Photon ---*/
	PgSetDrawBufferSize( 8*1024 );

	/*--- How many levels are available ? ---*/
	nb_levels = sizeof( conf_levels ) / sizeof( CONF_LEVEL );

	/*--- Frame update active ? ---*/
	is_drawFrame = FALSE;

	/*--- Freeze the game ? ---*/
	is_frozen = FALSE;

	/*--- .ini filename ---*/
	strcpy( ini_fname, "" );
	if ( getlogin() && ((pw = getpwnam( getlogin() )) != NULL) )
		{
		char	p[ PATH_MAX+1 ];
		int		l;
		strcpy( p, pw->pw_dir );
		if ( ((l=strlen(p)) >= 1) && (p[l-1]=='/') )
			p[l-1] = 0;
		sprintf( ini_fname, "%s/.photon/%s", p, CFG_FILENAME );
		}

	/*--- Testing level ? ---*/
	testing_level = -1;
	for ( n = 1; n < argc; n++ )
		if ( !strcmp( argv[ n ], "TestLevel#" ) )
			{
			testing_level = atoi( argv[ n+1 ] ) - 1;
			if ( (testing_level < 0) || (testing_level >= nb_levels) )
				testing_level = -1;
			break;
			}

	/*--- Set the size ---*/
	background = FALSE;
	set_min_size();

	/*--- Function signaled each 20msec to prepare the frame ---*/
	if ( (proxy_updateFrame = qnx_proxy_attach( 0, 0, 0, -1 )) == -1 )
		{
		fprintf( stderr, "No more proxy available.\n" );
		exit( -1 );
		}
	sigevnt.sigev_signo = -proxy_updateFrame;
	if ( (timer_updateFrame = timer_create( CLOCK_REALTIME, &sigevnt )) == -1 ) 
		{
		fprintf( stderr, "No more timer available.\n" );
		exit( -1 );
		}

	/*--- Proxy to be signaled when the cannon is dead ---*/
	if ( (proxy_cannonIsDead = qnx_proxy_attach( 0, 0, 0, getprio( 0 ) + 1 )) == -1 )
		{
		fprintf( stderr, "No more proxy available.\n" );
		exit( -1 );
		}

	/*--- Open the main window ---*/
	ApCreateModule( ABM_phreshold, NULL, NULL );

	/*--- Some global initializations (once, that's all) ---*/
	init_first();

	/*--- Create the process for drawing frame ---*/
	pid_kbd = -1;
	create_kbd_task();

	/*--- Get system infos ---*/
	PtQuerySystemInfo( ABW_phreshold, &sysinfo );

	/*--- Compute how many frames the current system is able to draw ---*/
	rid =  PtWidgetRid( ABW_phreshold );
	memset( work_frame, 0, frame_size );
	for ( t_start = clock(), nb_frames = 0;
		  (clock() - t_start) < (CLOCKS_PER_SEC * 1);
		  nb_frames++ )
		{
		PgSetRegion( rid );
		PgDrawImagemx( work_frame, Pg_IMAGE_PALETTE_BYTE, &frame_pos, &frame_dim, frame_width, 0 );
		PgFlush();
		}

	/*--- Max nb of frames per second ---*/
	max_nbframes_persec = min( 50, nb_frames );

	/*--- Compute how many frames the current system is able to draw ---*/
	setString( ABW_phreshold_msg, 
		"System capability: %d frames/sec\n%d bits / color", nb_frames, sysinfo.gfx.color_bits );
	if ( sysinfo.gfx.color_bits != 8 )
		message( "256 colors",
			"In order to play the game, the graphics driver\n"
			"MUST be configured to use 256 colors.\n"
			"Please run the mode-switcher...\n[%d]", sysinfo.gfx.color_bits );
	if ( nb_frames < 75 )
		message( "Not playable", 
			"The system capability seems that this\n"
			"game is not really playable.\n"
			"Try an other graphics driver (Pg.flat if possible)" );

	/*--- Verify ticksize (1msec better) ---*/
	qnx_osinfo( 0, &osdata );
	if ( osdata.tick_size > 5000 )
		message( "Warning: Ticksize", "Ticksize should be set at least to 5msec for a smooth game." );

//	extend_phreshold_size();

//	PtAppAddWorkProc( NULL, main_loop, NULL );

	/*--- Link to the shared memory which hold the keys state ---*/
	for ( cpt = 0; cpt < 10; cpt++ )
		{
		if ( (fd = shm_open( KEYS_SHNAME, O_RDWR, 0777 )) != -1 )
			break;
		delay( 100 );
		}
	if ( cpt == 10 )
		{
		fprintf( stderr, "Unable to attach shared memory [%d].\n", errno );
		exit( -1 );
		}
	if ( (keys_state = mmap( 0, sizeof( KEYS_SHARED ), PROT_READ, MAP_SHARED, fd, 0 )) == (void*)-1 )
		{
		fprintf( stderr, "Unable to attach shared memory [%d].\n", errno );
		exit( -1 );
		}

	/*--- Set priority ---*/
	setprio( 0, 10 );

	/*--- Testing mode ? ---*/
	if ( testing_level != -1 )
		{

		/*--- Begin to play ---*/
		start_play();
		
		}	// if

	/*--- Loop until user quits application ---*/
	if ( (event = malloc( sizeof( PhEvent_t ) + 1000 )) == NULL )
		{
		fprintf( stderr, "No memory !\n" );
		exit( -1 );
		}
	PhEventArm();
	for (;;)
		{
		pid = Receive( 0, NULL, 0 );
		if ( pid == proxy_updateFrame )
			drawFrame();
		else if ( pid  == proxy_cannonIsDead )
			cannon_is_dead();
		else
			if ( PhEventRead( pid, event, sizeof( PhEvent_t ) + 1000 ) == Ph_EVENT_MSG )
				PtEventHandler( event );
			else
				PhEventArm();
		}	// for (;;)
	
//	return( Pt_CONTINUE );

	}	// phreshold_init


void refresh_frame( const BOOL on )
	{
	struct itimerspec	timerSpec;

	if ( on )
		{

		/*--- Fired each 20msec (50 frames/sec) ---*/
		timerSpec.it_value.tv_sec     = 0;			
		timerSpec.it_value.tv_nsec    = 20000000;	// 20,000,000 nsec before initial shoot
		timerSpec.it_interval.tv_sec  = 0;
		timerSpec.it_interval.tv_nsec = 20000000;	// Repetitive
		timer_settime( timer_updateFrame, 0, &timerSpec, NULL );

		/*--- Set the palette ---*/
		PgSetPalette( palette, 0, 0, 256, Pg_PALSET_SOFT, crc_palette );
		PgFlush();

		/*--- Frame update active ? ---*/
		is_drawFrame = TRUE;

		}
	else
		{

		/*--- Stop timer ---*/
		timerSpec.it_value.tv_sec     = 
		timerSpec.it_value.tv_nsec    = 
		timerSpec.it_interval.tv_sec  = 
		timerSpec.it_interval.tv_nsec = 0;
		timer_settime( timer_updateFrame, 0, &timerSpec, NULL );

		/*--- Frame update active ? ---*/
		is_drawFrame = FALSE;

		}

	/*--- Flush all pending proxy ---*/
	while (Creceive( proxy_updateFrame, NULL, 0 ) == proxy_updateFrame );

	}	// refresh_frame


int
phreshold_setup( PtWidget_t *link_instance, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{

	/* eliminate 'unreferenced' warnings */
	link_instance = link_instance, apinfo = apinfo, cbinfo = cbinfo;

	/*--- Set the Position/Size of the zoom window if given in the .ini file ---*/
	readandset_cfg_szpos( "Main window", ABW_phreshold );

	return( Pt_CONTINUE );

	}	// phreshold_setup


int
phreshold_destroyed( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	/*--- Upcate the .ini file (configuration file) ---*/
	update_config_file();

	/*--- Des-allocate shared memory for fast draw of life ---*/
	PgShmemDestroy( work_frame );

#if LOGTOTMP==1
	slog( "exit" );
#endif

	return( Pt_CONTINUE );

	}	// phreshold_destroyed


void draw_score( void )
	{
	
	setString( ABW_phreshold_score, "%d", score );
	PgFlush();
	
	}	// draw_score


void drawFrame( void )
	{
static	int			first = TRUE;
static	PhEvent_t	*event = NULL;
static	PhRid_t		rid;

	/*--- Create a semaphore to control the use of the frame (shared memory) ---*/
	if ( first )
		{
		event = (PhEvent_t*)malloc( sizeof( PhEvent_t ) + 1000 );
		rid = PtWidgetRid( ABW_phreshold );
		first = FALSE;
		}

	/*--- Cannon is dead ? ---*/
	if ( cannon.is_hit && !cannon.hitcnt )
		{
		refresh_frame( FALSE );
		Trigger( proxy_cannonIsDead );
		return;
		}

	/*--- Prepare the frame: get the background, w/o objects ---*/
	memmove( work_frame, background, frame_size );

	/*--- Draw the stars ---*/
	draw_stars();

	/*--- Move the ships armada ---*/
	move_ships();

	if ( is_starting_cannon )
		{
		
		move_starting_cannon();

		}
	else
		{

		/*--- Move the shells ---*/
		move_shells();

		/*--- Draw the cannon ---*/
		draw_cannon();

		/*--- Move the missile, if any ---*/
		draw_missile();

		/*--- Update the fuel level ---*/
		update_fuel();
	
		/*--- Update the laser level ---*/
		update_laser();
		
		}	// if

	PgSetRegion( rid );
	PgDrawImagemx( work_frame, Pg_IMAGE_PALETTE_BYTE, &frame_pos, &frame_dim, frame_width, 0 );
	PgFlush();

	}	// drawFrame


void freeze( void )
	{

	if ( is_drawFrame && !is_frozen )
		{
		refresh_frame( FALSE );
		when_frozen = time( NULL );
		is_frozen = TRUE;
		}	// if
	
	}	// freeze


void unfreeze( void )
	{
	SHIP	*ship;
	int		n;
	int		delta;
	
	if ( is_frozen )
		{

		delta = (int)difftime( time( NULL ), when_frozen );
		for ( ship = ships, n = 0; n < currlevel.nbships; n++, ship++ )
			{
			if ( ship->when_nonvisible )
				ship->when_nonvisible -= delta;
			}	// for (ship, n)

		is_frozen = FALSE;
		refresh_frame( TRUE );

		}	// if
	
	}	// unfreeze


int
phreshold_window( PtWidget_t *widget, ApInfo_t *apinfo, PtCallbackInfo_t *cbinfo )

	{
	PhWindowEvent_t		*data = (PhWindowEvent_t *)cbinfo->cbdata;

	/* eliminate 'unreferenced' warnings */
	widget = widget, apinfo = apinfo, cbinfo = cbinfo;

	if ( data->event_state == Ph_WM_EVSTATE_FOCUSLOST )
		{
		freeze();
		}
	else if ( data->event_state == Ph_WM_EVSTATE_FOCUS )
		{
		unfreeze();
		}

	return( Pt_CONTINUE );

	}	// phreshold_window
